import 'dart:collection';
import 'package:dart_style/dart_style.dart';
import 'helpers.dart';
import 'syntax.dart';

// ignore: import_of_legacy_library_into_null_safe
import 'package:json_ast/json_ast.dart' show parse, Settings, Node;

class DartCode extends WithWarning<String> {
  DartCode(String result, List<Warning> warnings) : super(result, warnings);

  String get code => this.result;
}

/// A Hint is a user type correction.
class Hint {
  final String path;
  final String type;

  Hint(this.path, this.type);
}

class ModelGenerator {
  final String _rootClassName;
  final bool _privateFields;
  List<ClassDefinition> allClasses = [];
  final Map<String, String> sameClassMapping = HashMap<String, String>();
  late List<Hint> hints;

  ModelGenerator(this._rootClassName, [this._privateFields = false, hints]) {
    if (hints != null) {
      this.hints = hints;
    } else {
      this.hints = [];
    }
  }

  Hint? _hintForPath(String path) {
    Hint? hint;
    try {
      hint = this.hints.firstWhere((h) => h.path == path);
    } catch (e) {
      hint = null;
    }
    return hint;
  }

  List<Warning> _generateClassDefinition(String className,
      dynamic jsonRawDynamicData, String path, Node? astNode) {
    List<Warning> warnings = [];
    if (jsonRawDynamicData is List) {
      // if first element is an array, start in the first element.
      final node = navigateNode(astNode, '0');
      _generateClassDefinition(className, jsonRawDynamicData[0], path, node);
    } else {
      final Map<dynamic, dynamic>? jsonRawData = jsonRawDynamicData;
      if (jsonRawData != null) {
        ClassDefinition classDefinition =
            ClassDefinition(className, _privateFields);
        final keys = jsonRawData.keys;
        keys.forEach((key) {
          TypeDefinition typeDef;
          final hint = _hintForPath('$path/$key');
          final node = navigateNode(astNode, key);
          if (hint != null) {
            typeDef = TypeDefinition(hint.type, astNode: node);
          } else {
            typeDef = TypeDefinition.fromDynamic(jsonRawData[key], node);
          }
          if (typeDef.name == 'Class') {
            // 处理特殊命名
            if (["list", "map", "function"].indexOf(key) >= 0) {
              typeDef.name = "Custom${camelCase(key)}";
            } else {
              typeDef.name = camelCase(key);
            }
          }
          if (typeDef.name == 'List' && typeDef.subtype == 'Null') {
            warnings.add(newEmptyListWarn('$path/$key'));
          }
          if (typeDef.subtype != null && typeDef.subtype == 'Class') {
            typeDef.subtype = camelCase(key);
          }
          if (typeDef.isAmbiguous!) {
            warnings.add(newAmbiguousListWarn('$path/$key'));
          }
          classDefinition.addField(key, typeDef);
        });
        ClassDefinition? similarClass;
        try {
          similarClass = allClasses.firstWhere((cd) => cd == classDefinition);
        } catch (e) {
          similarClass = null;
        }
        if (similarClass != null) {
          final similarClassName = similarClass.name;
          final currentClassName = classDefinition.name;
          sameClassMapping[currentClassName] = similarClassName;
        } else {
          allClasses.add(classDefinition);
        }
        final dependencies = classDefinition.dependencies;
        dependencies.forEach((dependency) {
          List<Warning>? warns;
          if (dependency.typeDef.name == 'List') {
            // only generate dependency class if the array is not empty
            if (jsonRawData[dependency.name].length > 0) {
              // when list has ambiguous values, take the first one, otherwise merge all objects
              // into a single one
              dynamic toAnalyze;
              if (!dependency.typeDef.isAmbiguous!) {
                WithWarning<Map> mergeWithWarning = mergeObjectList(
                    jsonRawData[dependency.name], '$path/${dependency.name}');
                toAnalyze = mergeWithWarning.result;
                warnings.addAll(mergeWithWarning.warnings);
              } else {
                toAnalyze = jsonRawData[dependency.name][0];
              }
              final node = navigateNode(astNode, dependency.name);
              warns = _generateClassDefinition(dependency.className, toAnalyze,
                  '$path/${dependency.name}', node);
            }
          } else {
            final node = navigateNode(astNode, dependency.name);
            if (["CustomList", "CustomMap", "CustomFunction"]
                    .indexOf(dependency.typeDef.name) >=
                0) {
              warns = _generateClassDefinition(
                  dependency.typeDef.name,
                  jsonRawData[dependency.name],
                  '$path/${dependency.name}',
                  node);
            } else {
              warns = _generateClassDefinition(
                  dependency.className,
                  jsonRawData[dependency.name],
                  '$path/${dependency.name}',
                  node);
            }
          }
          if (warns != null) {
            warnings.addAll(warns);
          }
        });
      }
    }
    return warnings;
  }

  /// generateUnsafeDart will generate all classes and append one after another
  /// in a single string. The [rawJson] param is assumed to be a properly
  /// formatted JSON string. The dart code is not validated so invalid dart code
  /// might be returned
  DartCode generateUnsafeDart(String rawJson) {
    final jsonRawData = decodeJSON(rawJson);
    final astNode = parse(rawJson, Settings());
    List<Warning> warnings =
        _generateClassDefinition(_rootClassName, jsonRawData, "", astNode);
    // after generating all classes, replace the omited similar classes.
    allClasses.forEach((c) {
      final fieldsKeys = c.fields.keys;
      fieldsKeys.forEach((f) {
        final typeForField = c.fields[f];
        if (sameClassMapping.containsKey(typeForField!.name)) {
          c.fields[f]!.name = sameClassMapping[typeForField.name]!;
        }
      });
    });
    return DartCode(allClasses.map((c) => c.toString()).join('\n'), warnings);
  }

  /// generateDartClasses will generate all classes and append one after another
  /// in a single string. The [rawJson] param is assumed to be a properly
  /// formatted JSON string. If the generated dart is invalid it will throw an error.
  DartCode generateDartClasses(String rawJson) {
    final unsafeDartCode = generateUnsafeDart(rawJson);
    final formatter = DartFormatter();
    return DartCode(
        formatter.format(unsafeDartCode.code), unsafeDartCode.warnings);
  }
}
